import typing
from functools import cached_property
from typing import Iterable, Literal, Optional

import pandas as pd

from .._io import loads_pickle
from .._models import BaseModel, create_model
from .._normalize import NormalizeColumns
from .._settings import (
    check_datasetdir_exists,
    check_dynamicdir_exists,
    format_into_dataframe,
    settings,
)
from ..species import Species as SP
from ._query import Biomart, Mygene

_IDs = Literal["ensembl.gene_id", "entrez.gene_id"]
_HGNC = "http://ftp.ebi.ac.uk/pub/databases/genenames/hgnc/tsv/hgnc_complete_set.txt"

GeneData = create_model("GeneData", __module__=__name__)


class Entry(BaseModel):
    hgnc_symbol: str
    hgnc_id: str
    name: str
    locus_group: str
    alias_symbol: str
    location: str
    entrez_gene_id: str
    ensembl_gene_id: str
    uniprot_ids: str


class Gene:
    """Gene.

    Biotypes: https://useast.ensembl.org/info/genome/genebuild/biotypes.html
    Gene Naming: https://useast.ensembl.org/info/genome/genebuild/gene_names.html

    """

    def __init__(self, species="human", biomart=False):
        self._species = SP(common_name=species)
        self._dataclasspath = (
            settings.dynamicdir / f"genedata_{self.species.std_name}.pkl"
        )
        self._ref = None
        self._biomart = biomart

    @property
    def species(self):
        """Bionty.species()."""
        return self._species

    @property
    def std_id(self):
        """The standardized symbol attribute name."""
        STD_ID_DICT = {"human": "hgnc_symbol", "mouse": "mgi_symbol"}
        return STD_ID_DICT[self.species.std_name]

    @property
    def fields(self):
        ATTR_DICT = {"human": ["hgnc_id", "hgnc_symbol"], "mouse": ["mgi_symbol"]}
        return list(typing.get_args(_IDs)) + ATTR_DICT[self.species.std_name]

    @property
    def dataclasspath(self):
        """Path to the picked dataclass."""
        return self._dataclasspath

    @property
    def biomart(self):
        """Whether to pull reference via the biomart API."""
        return self._biomart

    @cached_property
    def reference(self):
        """Gene reference table."""
        self._pull_ref()
        return self._ref

    @cached_property
    def dataclass(self):
        return self._load_dataclass()

    @check_dynamicdir_exists
    def _load_dataclass(self):
        """Pydantic data class of genes."""
        if not self.dataclasspath.exists():
            import pickle

            from .._io import write_pickle

            model = self._create_data_model()
            write_pickle(pickle.dumps(model()), self.dataclasspath)

        return loads_pickle(self.dataclasspath)

    def _create_data_model(self):
        """Create the gene data model with pydantic."""
        df = self.reference.fillna("")
        df.columns = df.columns.str.replace(".", "_", regex=True)
        df["entrez_gene_id"] = (
            df["entrez_gene_id"].astype(str).str.replace(".0", "", regex=False)
        )
        df = df.loc[:, df.columns.isin(Entry.__annotations__.keys())].copy()
        for i in df.index:
            entry = {}
            entry.update({col: df.loc[i][col] for col in df.columns})
            GeneData.add_fields(**{df.loc[i][self.std_id]: (Entry, Entry(**entry))})

        return GeneData

    @format_into_dataframe
    def standardize(
        self,
        data,
        id_type: Optional[_IDs] = None,
        new_index: bool = True,
        _reformat: bool = False,
    ):
        """Index a dataframe with the official gene symbols from HGNC.

        Args:
            data: A list of gene symbols to be standardized
                If dataframe, will take the index
            id_type: Default is to consider input as gene symbols and alias
            new_index:
                If True, set the standardized symbols as the index
                    - unmapped will remain the original index
                    - original index stored in the `index_orig` column

                If False, write to the `standardized_symbol` column

        Returns:
            Replaces the DataFrame mappable index with the standardized symbols
            Adds a `std_id` column
            The original index is stored in the `index_orig` column
        """
        if id_type is None:
            mapped_dict = self._standardize_symbol(df=data)
        else:
            mapped_dict = self.search(
                data.index, id_type_from=id_type, id_type_to=self.std_id
            )

        data["std_id"] = data.index.map(mapped_dict)
        if new_index:
            data["index_orig"] = data.index
            data.index = data["std_id"].fillna(data["index_orig"])
            data.index.name = None

        if _reformat:
            return data

    def search(
        self,
        genes: Iterable[str],
        id_type_from="ensembl_gene_id",
        id_type_to=None,
    ):
        """Search among fields that are in the `.reference` table.

        Args:
            genes: Input list
            id_type_from: ID type of the input list, see `.attributes`
            id_type_to: ID type to convert into
                Default is the `.std_id`

        Returns:
            a dict of mapped ids
        """
        # default if to convert tp the standardized id
        if id_type_to is None:
            id_type_to = self.std_id

        # get mappings from the reference table
        df = self.reference.reset_index().set_index(id_type_from)[[id_type_to]].copy()

        return df[df.index.isin(genes)].to_dict()[id_type_to]

    def _pull_ref(self):
        """Pulling gene reference table.

        If biomart, pull the reference table from biomart
        If not, pull the reference table from HGNC directly
        """
        if self.biomart:
            ref = Biomart().get_gene_ensembl(species=self.species.std_name)
            if "entrezgene_id" in ref.columns:
                ref["entrezgene_id"] = (
                    ref["entrezgene_id"]
                    .fillna(0)
                    .astype(int)
                    .astype(object)
                    .where(ref["entrezgene_id"].notnull())
                )
            self._ref = ref
        else:
            self._ref = self.hgnc(species=self.species.std_name)

    def _standardize_symbol(
        self,
        df: pd.DataFrame,
    ):
        """Standardize gene symbols/aliases to symbol from `.reference` table.

        Args:
            df: A dataframe with index being the column to be standardized
            species: 'human'

        Returns:
            a dict with the standardized symbols
        """
        # 1. Mapping from symbol to hgnc_id using .hgnc table
        mapped_dict = self.search(df.index, "hgnc_symbol", "hgnc_id")
        mapped_dict.update({k: k for k in mapped_dict.keys()})

        # 2. For not mapped symbols, map through alias
        notmapped = df[~df.index.isin(mapped_dict.keys())].copy()
        if notmapped.shape[0] > 0:
            mg = Mygene()
            res = mg.query(
                notmapped.index, scopes="symbol,alias", species=self.species.std_name
            )
            mapped_dict.update(self._cleanup_mygene_returns(res))

        return mapped_dict

    def _cleanup_mygene_returns(self, res: pd.DataFrame, unique_col="hgnc_id"):
        """Clean up duplicates and NAs from the mygene returns.

        Args:
            res: Returned dataframe from `.mg.query`
            unique_col: Unique identifier column

        Returns:
            a dict with only uniquely mapped IDs
        """
        mapped_dict = {}

        # drop columns without mapped unique IDs (HGNC)
        df = res.dropna(subset=unique_col)

        # for unique results, use returned HGNC IDs to get symbols from .hgnc
        udf = df[~df.index.duplicated(keep=False)].copy()
        udf["std_id"] = udf["hgnc_id"].map(
            self.search(udf["hgnc_id"], "hgnc_id", "hgnc_symbol")
        )
        mapped_dict.update(udf[["std_id"]].to_dict()["std_id"])

        # TODO: if the same HGNC ID is mapped to multiple inputs?
        if df[unique_col].duplicated().sum() > 0:
            pass

        # if a query is mapped to multiple HGNC IDs, do the reverse mapping from .hgnc
        # keep the shortest symbol as readthrough transcripts or pseudogenes are longer
        if df.index.duplicated().sum() > 0:
            dups = df[df.index.duplicated(keep=False)].copy()
            for dup in dups.index.unique():
                hids = dups[dups.index == dup][unique_col].tolist()
                d = self.search(hids, "hgnc_id", "hgnc_symbol")
                mapped_dict[dup] = pd.DataFrame.from_dict(d, orient="index")[0].min()

        return mapped_dict

    @check_datasetdir_exists
    def hgnc(self, species="human"):
        """HGNC symbol from the HUGO Gene Nomenclature Committee."""
        if species != "human":
            raise AssertionError("HGNC is only for human!")

        filepath = settings.datasetdir / "hgnc_complete_set.txt"
        if not filepath.exists():
            from urllib.request import urlretrieve

            urlretrieve(_HGNC, filepath)
        df = pd.read_csv(
            filepath,
            sep="\t",
            index_col=0,
            low_memory=False,  # If True, gets DtypeWarning
            verbose=False,
        )
        df = df.reset_index().copy()
        NormalizeColumns.gene(df, species=species)

        return df



import asyncio
import sys
from typing import Any

import httpx
import nest_asyncio
import requests  # type: ignore
from syncer import sync

nest_asyncio.apply()  # Fixes the issue with iPython compatibility


async def _async_get(client, url, content_type="application/json"):
    resp = await client.get(url, headers={"content-type": content_type})

    if resp.headers["content-type"].find("application/json") > -1:
        return resp.json()
    else:
        return resp.text


@sync
async def get_request_async(
    server_request: str, terms, content_type="application/json"
):
    async with httpx.AsyncClient() as client:

        tasks = []

        for term in terms:
            url = f"{server_request}{term}"
            tasks.append(
                asyncio.ensure_future(
                    _async_get(client, url, content_type=content_type)
                )
            )

        resps = await asyncio.gather(*tasks)

    return resps


def get_request(server, request, content_type="application/json", **kwds) -> Any:
    """Fetch an endpoint from the server.

    Allow overriding of default content-type
    """
    r = requests.get(server + request, headers={"Accept": content_type}, **kwds)

    if not r.ok:
        r.raise_for_status()
        sys.exit()

    if content_type == "application/json":
        return r.json()
    else:
        return r.text


def post_request(server, request, data, content_type="application/json", **kwds) -> Any:
    """POST requests."""
    r = requests.post(
        server + request, headers={"Content-Type": content_type}, data=data, **kwds
    )

    if not r.ok:
        r.raise_for_status()
        sys.exit()

    if content_type == "application/json":
        return r.json()
    else:
        return r.text



import io
import urllib.parse
import urllib.request

import pandas as pd

from ..species import Species as SP


class Uniprot:
    """Wrapper of the Uniprot REST APIs.

    See: https://www.uniprot.org/help/api_idmapping
    """

    _URL = "https://www.uniprot.org/uploadlists/"

    def __init__(self) -> None:
        pass

    def query(
        self,
        prots,
        id_type_from="UNIPROT_ID",
        id_type_to="ENSEMBL_ID",
        columns=None,
        species="human",
    ):

        # replace UNIPROT_ID with ACC.
        colnames = (
            columns.split(",") + [id_type_from, id_type_to]
            if columns is not None
            else [id_type_from, id_type_to]
        )
        id_type_from = "ACC" if id_type_from == "UNIPROT_ID" else id_type_from
        id_type_to = "ACC" if id_type_to == "UNIPROT_ID" else id_type_to

        # taxon id of species
        taxon_id = SP(common_name=species).search("taxon_id")

        # set up params
        params = {
            "from": id_type_from,
            "to": id_type_to,
            "columns": columns,
            "format": "tab",
            "query": " ".join([i for i in prots]),
            "taxon": taxon_id,
        }

        # query uniprot
        data = urllib.parse.urlencode(params)
        data = data.encode("utf-8")
        req = urllib.request.Request(self._URL, data)
        with urllib.request.urlopen(req) as f:
            response = f.read()

        # format results into a dataframe
        data = response.decode("utf-8")
        df = pd.read_csv(io.StringIO(data), sep="\t", header=0)
        df.columns = colnames

        return df



from typing import Any, Dict, Optional

from pydantic import BaseModel as PydanticBaseModel
from pydantic import create_model as pydantic_create_model
from pydantic.fields import ModelField


class BaseModel(PydanticBaseModel):
    @classmethod
    def add_fields(cls, **field_definitions: Any):
        """Dynamically adding fields inplace.

        From: https://github.com/samuelcolvin/pydantic/issues/1937
        """
        new_fields: Dict[str, ModelField] = {}
        new_annotations: Dict[str, Optional[type]] = {}

        for f_name, f_def in field_definitions.items():
            if isinstance(f_def, tuple):
                try:
                    f_annotation, f_value = f_def
                except ValueError as e:
                    raise Exception(
                        "field definitions should either be a tuple of (<type>,"
                        " <default>) or just a default value, unfortunately this means"
                        " tuples as default values are not allowed"
                    ) from e
            else:
                f_annotation, f_value = None, f_def

            if f_annotation:
                new_annotations[f_name] = f_annotation

            new_fields[f_name] = ModelField.infer(
                name=f_name,
                value=f_value,
                annotation=f_annotation,
                class_validators=None,
                config=cls.__config__,
            )

        cls.__fields__.update(new_fields)
        cls.__annotations__.update(new_annotations)


class Entity(BaseModel):
    name: str  # this is the value of the standardized id
    std_id: str  # which field contains the standardized ids


def create_model(__model_name: str, *, __base__=BaseModel, **kwargs):
    return pydantic_create_model(
        __model_name=__model_name, __base__=BaseModel, **kwargs
    )



import pickle
from pathlib import Path
from typing import Union

import orjson
from typeguard import typechecked


@typechecked
def read_pickle(filename: Union[str, Path]):
    with open(filename, "rb") as handle:
        return pickle.load(handle)


def loads_pickle(filename: Union[str, Path]):
    with open(filename, "rb") as handle:
        r = pickle.load(handle)
    return pickle.loads(r)


@typechecked
def write_pickle(data, filename: Union[str, Path]) -> None:
    with open(filename, "wb") as handle:
        pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)


@typechecked
def read_json(filename: Union[str, Path]):
    with open(filename, "rb") as f:
        return orjson.loads(f.read())


@typechecked
def write_json(data, filename: Union[str, Path]):
    with open(filename, "wb") as f:
        f.write(orjson.dumps(data))



from .._ontology import Ontology
from .._urls import OBO_CL_OWL


class CellType(Ontology):
    """Cell type bioentity.

    Edits of terms are coordinated and reviewed on:
    https://github.com/obophenotype/cell-ontology
    """

    def __init__(self, base_iri=OBO_CL_OWL, load: bool = True) -> None:
        super().__init__(base_iri=base_iri, load=load)



import xmltodict

from ._httpx import get_request, get_request_async, post_request
from ._urls import ENSEMBL_REST, ENSEMBL_REST_EXT


class EnsemblREST:
    """Queries via the Ensembl REST APIs."""

    def __init__(self) -> None:
        self._server = ENSEMBL_REST

    @property
    def server(self):
        """ENSEMBL_REST."""
        return self._server

    def _config_data(self, ids, label):
        ids = str(ids).replace("'", '"')
        return f'{{ "{label}" : {ids} }}'

    def species_info(self, return_raw=False):
        """ENSEMBL_REST_EXT.SPECIES_INFO."""
        ext = ENSEMBL_REST_EXT.SPECIES_INFO
        res = get_request(self.server, ext, "text/xml")
        if return_raw:
            return res
        else:
            return xmltodict.parse(res)["opt"]["data"]["species"]

    def xref(self, ids, **kwargs):
        """Retrieve external references of Ensembl ids.

        See https://rest.ensembl.org/documentation/info/xref_id
        """
        if isinstance(ids, str):
            ext = f"{ENSEMBL_REST_EXT.XREFS_ID}{ids}?"
            res = get_request(self.server, ext, **kwargs)
        else:
            res = get_request_async(self.server + ENSEMBL_REST_EXT.XREFS_ID, ids)
        return res

    def archive_ids(self, ids):
        """Retrieve the latest version for a set of identifiers."""
        ext = ENSEMBL_REST_EXT.ARCHIVE_IDS
        res = post_request(self.server, ext, data=self._config_data(ids, "id"))
        return res

    def lookup_ids(self, ids, **kwargs):
        """Find the species and database for several identifiers.

        See https://rest.ensembl.org/documentation/info/lookup_post
        """
        ext = ENSEMBL_REST_EXT.LOOKUP_IDS
        res = post_request(
            self.server, ext, data=self._config_data(ids, "id"), **kwargs
        )
        return res

    def lookup_symbols(self, symbols, species="homo_sapiens", **kwargs):
        """Find the species and database for symbols in a linked external database."""
        ext = f"{ENSEMBL_REST_EXT.LOOKUP_SYMBOLS}{species}"
        res = post_request(self.server, ext, data=self._config_data(symbols), **kwargs)
        return res

    def seq_ids(self, ids, **kwargs):
        """Request multiple types of sequence by a stable identifier list."""
        ext = ENSEMBL_REST_EXT.SEQ_IDS
        res = post_request(self.server, ext, data=self._config_data(ids), **kwargs)
        return res



from functools import cached_property

from .._io import loads_pickle
from .._models import create_model
from .._ontology import Ontology
from .._settings import check_dynamicdir_exists, settings
from .._urls import OBO_MONDO_OWL

DiseaseData = create_model("DiseaseData", __module__=__name__)


class Disease(Ontology):
    """Disease bioentity.

    Edits of terms are coordinated and reviewed on:
    https://github.com/monarch-initiative/mondo
    """

    def __init__(self) -> None:
        self._dataclasspath = settings.dynamicdir / "diseasedata.pkl"
        if not self.dataclasspath.exists():
            super().__init__(base_iri=OBO_MONDO_OWL, load=True)

    @property
    def dataclasspath(self):
        """Path to the picked dataclass."""
        return self._dataclasspath

    @cached_property
    def onto_dict(self) -> dict:
        """Keyed by name, valued by label."""
        return {
            v.name: v.label[0]
            for k, v in self.classes.items()
            if k.startswith("MONDO") & len(v.label) > 0
        }

    @cached_property
    def dataclass(self):
        return self._load_dataclass()

    @check_dynamicdir_exists
    def _load_dataclass(self):
        """Pydantic data class of diseases."""
        if not self.dataclasspath.exists():
            import pickle

            from .._io import write_pickle

            DiseaseData.add_fields(**self.onto_dict)
            write_pickle(pickle.dumps(DiseaseData()), self.dataclasspath)

        return loads_pickle(self.dataclasspath)



from functools import wraps
from pathlib import Path
from typing import Union

import pandas as pd

ROOT_DIR = Path(__file__).parent.resolve()


def check_datasetdir_exists(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        settings.datasetdir.mkdir(exist_ok=True)
        return f(*args, **kwargs)

    return wrapper


def check_dynamicdir_exists(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        settings.dynamicdir.mkdir(exist_ok=True)
        return f(*args, **kwargs)

    return wrapper


def format_into_dataframe(f):
    @wraps(f)
    def dataframe(*args, **kwargs) -> pd.DataFrame:
        # Check if the first argument is self
        idx = 0 if _is_function(args[0]) else 1

        if isinstance(args[idx], pd.DataFrame):
            df = args[idx]
            reformat = False
        else:
            df = pd.DataFrame(index=[d for d in args[idx]])
            reformat = True

        args_new = list(args)
        args_new[idx] = df
        return f(*args_new, _reformat=reformat, **kwargs)

    return dataframe


def _is_function(func) -> bool:
    """Checks if input is a function rather than an instance of class.

    TODO: Is this hacky?
    """
    return hasattr(func, "__name__")


@check_dynamicdir_exists
def dump_dataclass_as_private_module(obj):
    """Save the dataclass as python code."""
    header = f"from pydantic import BaseModel\n\n\nclass {obj.__name__}(BaseModel):\n"
    with open(settings.dynamicdir / f"{obj.__name__.lower()}.py", "w") as f:
        f.write(header)
        for key, value in obj().__fields__.items():
            f.write(f"    {key} : {value.type_.__name__} = {value.default!r}\n")


class Settings:
    def __init__(
        self,
        datasetdir: Union[str, Path] = ROOT_DIR / "data/",
        dynamicdir: Union[str, Path] = ROOT_DIR / "_dynamic/",
    ):
        self.datasetdir = datasetdir
        self.dynamicdir = dynamicdir

    @property
    def datasetdir(self):
        """Directory for datasets."""
        return self._datasetdir

    @datasetdir.setter
    def datasetdir(self, datasetdir: Union[str, Path]):
        self._datasetdir = Path(datasetdir).resolve()

    @property
    def dynamicdir(self):
        """Directory for datasets."""
        return self._dynamicdir

    @dynamicdir.setter
    def dynamicdir(self, dynamicdir: Union[str, Path]):
        self._dynamicdir = Path(dynamicdir).resolve()


settings = Settings()



from argparse import Namespace

# Ensembl REST server
ENSEMBL_REST = "http://rest.ensembl.org"
ENSEMBL_REST_EXT = Namespace(
    **{
        "SPECIES_INFO": "/info/species?",  # all species info
        "ARCHIVE_IDS": "/archive/id",  # retrieves the latest version of ids
        "XREFS_ID": "/xrefs/id/",
        "LOOKUP_IDS": "/lookup/id",
        "LOOKUP_SYMBOLS": "/lookup/symbol/",
        "SEQ_IDS": "/sequence/id",
    }
)

# HGNC REST server
"""See searchable fields here: https://www.genenames.org/help/rest/"""
HGNC_REST = "http://rest.genenames.org/"
HGNC_REST_EXT = Namespace(INFO="info", FETCH="fetch/", SEARCH="search/")

# The Open Biological and Biomedical Ontology (OBO) Foundry
"""Website: https://obofoundry.org/"""
OBO = "http://purl.obolibrary.org/obo/"
OBO_CL = OBO + "cl/cl-simple.obo"
OBO_CL_OWL = OBO + "cl/cl-simple.owl"
OBO_MONDO_OWL = OBO + "mondo/mondo-base.owl"
OBO_UBERON_OWL = OBO + "uberon/uberon-base.owl"



import pandas as pd

GENE_COLUMNS = {
    "HGNC": "hgnc_id",
    "ensembl_gn_id": "ensembl.gene_id",
    "ensembl_gene_id": "ensembl.gene_id",
    "entrezgene_id": "entrez.gene_id",
    "entrez_id": "entrez.gene_id",
}


class NormalizeColumns:
    """Standardizing column names."""

    def __init__(self) -> None:
        pass

    @staticmethod
    def gene(df: pd.DataFrame, species=None):
        """Column names of gene tables.

        We try to adapt a naming system that is {database}.{id_type} when
        multipleids exist within that database.
        - e.g. hgnc_id is the only id in HGNC, therefore it's not using the .
        - e.g. ensembl can have ensembl.gene_id and ensembl.transcript_id
        """
        if species == "human":
            GENE_COLUMNS.update({"symbol": "hgnc_symbol"})
        df.rename(columns=GENE_COLUMNS, inplace=True)



"""Bionty: mapping and standardizing biological entities.

Import the package::

   import bionty as bt

This is the complete API reference:

Base models: base entity managers.

.. autosummary::
   :toctree: .

   Ontology

Entity classes: dynamic classes of entities.

.. autosummary::
   :toctree: .

   CellType
   Disease
   Gene
   Protein
   Species
   Tissue

Entity collections: static classes of entities.

.. autosummary::
   :toctree: .

   gene
   species
"""

from . import _version

__version__ = "0.1a3"


# dynamic classes
from .gene import Gene
from .protein import Protein
from .species import Species
from .celltype import CellType
from .disease import Disease
from .tissue import Tissue

# tools
from ._normalize import NormalizeColumns
from ._ontology import Ontology



import typing
from typing import Iterable, Literal, Optional

from ..species import Species as SP
from ._query import Uniprot

_IDs = Literal["UNIPROT_ID", "PDB_ID", "CHEMBL_ID", "DRUGBANK_ID", "ENSEMBL_PRO_ID"]


class Protein:
    """Protein."""

    def __init__(self, species="human"):
        self._species = SP(common_name=species)

    @property
    def species(self):
        """bionty.species()."""
        return self._species

    @property
    def std_id(self):
        """Standardized id."""
        return "UNIPROT_ID"

    @property
    def fields(self):
        """List all protein related fields."""
        return list(typing.get_args(_IDs))

    def search(
        self,
        prots: Iterable[str],
        id_type_from: Optional[_IDs],
        id_type_to: Optional[_IDs] = None,
    ):
        """Mapping between protein IDs.

        Args:
            prots: Input list
            id_type_from: ID type of the input list, see `.fields`
            id_type_to: ID type to convert into
                Default is the `.std_id`

        Returns:
            a dict of mapped ids
        """
        # default is to convert into std_id
        id_type_to = self.std_id if id_type_to is None else id_type_to

        # get mappings from the reference table
        res = Uniprot().query(
            prots,
            id_type_from=id_type_from,
            id_type_to=id_type_to,
            species=self.species.std_name,
        )
        df = res.reset_index().set_index(id_type_from)[[id_type_to]].copy()

        return df[df.index.isin(prots)].to_dict()[id_type_to]



import logging as logg
from functools import cached_property
from pathlib import Path
from typing import Iterable, Union

import pandas as pd

from ._settings import format_into_dataframe


class Ontology:
    """Ontology manager built on Owlready2.

    Args:
        base_iri: (Internationalized Resource Identifier) RDF/XML, OWL/XML or NTriples format # noqa
        load: Whether to load ontology
    """

    def __init__(self, base_iri: Union[str, Path], load: bool = True) -> None:
        from owlready2 import get_ontology

        if load:
            self._onto = get_ontology(base_iri).load()
        else:
            self._onto = get_ontology(base_iri)

    @property
    def onto(self):
        """owlready2 ontology object."""
        return self._onto

    @cached_property
    def onto_dict(self) -> dict:
        """Keyed by name, valued by label."""
        return {i.name: i.label[0] for i in self.onto.classes()}

    @cached_property
    def classes(self) -> dict:
        """Indexed classes, owlready2 ThingClass object."""
        return {i.name: i for i in self.onto.classes()}

    def search(self, data: Iterable[str], id=False):
        """Search in ontology labels.

        Args:
            data: search patterns
            id: whether to search by the id

        Returns:
            A list of ontology names
        """
        res = {}
        for d in data:
            res[d] = self.search_one(text=d)
        return res

    def search_one(self, text: str, id=False) -> dict:
        """Search in ontology labels one by one.

        Args:
            text: search pattern
            id: whether to search by the id

        Returns:
            A list of ontology names
        """
        if id:
            text = text.replace(":", "_")
            res = self.onto.search(iri=f"*{text}")
        else:
            res = self.onto.search(label=text)

        return {i.name: i.label[0] for i in res}

    @format_into_dataframe
    def standardize(self, terms: pd.DataFrame, _reformat=False):
        """Checks if the ontology names are valid and in use.

        Args:
            terms: ontology ids

        Returns:
            a dataframe
        """
        terms.index = terms.index.str.replace(":", "_")
        terms.index.name = "ontology_id"
        terms["name"] = ""
        nonstd = {}
        for term in terms.index:
            # Ensuring the format of the IDs
            if term in self.onto_dict.keys():
                label = self.onto_dict[term]
                terms.loc[term]["name"] = label
                if label.startswith("obsolete"):
                    nonstd[term] = label
            else:
                nonstd[term] = label

        if len(nonstd) > 0:
            logg.warn(
                "The following terms were found to be obsolete or non-exist! Please"
                f" search the correct term via `.search`! \n {nonstd}"
            )

        if _reformat:
            return terms



import io
from typing import Iterable

import pandas as pd
from biothings_client import get_client

from .._normalize import NormalizeColumns


class Mygene:
    """Wrapper of MyGene.info.

    See: https://docs.mygene.info/en/latest/index.html
    """

    def __init__(self) -> None:
        self._server = get_client("gene")

    @property
    def server(self):
        """MyGene.info."""
        return self._server

    def query(
        self,
        genes: Iterable[str],
        scopes="symbol",
        fields="HGNC,symbol",
        species="human",
        as_dataframe=True,
        verbose=False,
        **kwargs,
    ):
        """Get HGNC IDs from Mygene.

        Args:
            genes: Input list
            scopes: ID types of the input
            fields: ID type of the output
            species: species
            as_dataframe: Whether to return a data frame
            verbose: Whether to print logging
            **kwargs: see **kwargs of `biothings_client.MyGeneInfo().querymany()`

        Returns:
            a dataframe ('HGNC' column is reformatted to be 'hgnc_id')
        """
        # query via mygene
        res = self.server.querymany(
            qterms=genes,
            scopes=scopes,
            fields=fields,
            species=species,
            as_dataframe=as_dataframe,
            verbose=verbose,
            **kwargs,
        )

        # format HGNC IDs to match `hgnc_id` format ('HGNC:int')
        if "HGNC" in res.columns:
            res["HGNC"] = [
                f"HGNC:{i}" if isinstance(i, str) else i for i in res["HGNC"]
            ]
        NormalizeColumns.gene(res)

        return res


class Biomart:
    """Wrapper of Biomart python APIs, good for accessing Ensembl data.

    See: https://github.com/sebriois/biomart
    """

    def __init__(self) -> None:
        try:
            import biomart

            self._server = biomart.BiomartServer("http://uswest.ensembl.org/biomart")
            self._dataset = None
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Run `pip install biomart`")

    @property
    def server(self):
        """biomart.BiomartServer."""
        return self._server

    @property
    def databases(self):
        """Listing all databases."""
        return self._server.databases

    @property
    def datasets(self):
        """Listing all datasets."""
        return self._server.datasets

    @property
    def dataset(self):
        """A biomart.BiomartDataset."""
        return self._dataset

    def get_gene_ensembl(
        self,
        species="human",
        attributes=None,
        filters={},
        **kwargs,
    ):
        """Fetch the reference table of gene ensembl from biomart.

        Args:
            species: common name of species
            attributes: gene attributes from gene_ensembl datasets
            filters: see biomart.search()
            **kwargs: see biomart.search()
        """
        # database name
        from bionty.gene import Gene

        gn = Gene(species=species)
        sname = gn.species.search("short_name")
        self._dataset = self.datasets[f"{sname}_gene_ensembl"]

        # default is to get all the attributes
        attributes = gn.fields if attributes is None else attributes

        # Get the mapping between the attributes
        response = self.dataset.search(
            {"filters": filters, "attributes": attributes},
            **kwargs,
        )
        data = response.raw.data.decode("utf-8")

        # returns a dataframe
        df = pd.read_csv(io.StringIO(data), sep="\t", header=None)
        df.columns = attributes

        return df



from pathlib import Path

import nox

nox.options.reuse_existing_virtualenvs = True


@nox.session
def lint(session: nox.Session) -> None:
    session.install("pre-commit")
    session.run("pre-commit", "install")
    session.run("pre-commit", "run", "--all-files")


@nox.session(python=["3.9"])
def build(session):
    session.install(".[dev,test]")
    session.run(
        "pytest",
        "--nbmake",
        "--overwrite",
    )  # write output instead of capturing it (more verbose)
    prefix = "." if Path("./lndocs").exists() else ".."
    session.install(f"{prefix}/lndocs")
    session.run("lndocs")



import sys

from loguru import logger

# all times in UTC
logger.configure(
    handlers=[
        dict(
            sink=sys.stdout,
            format="{time:YYYY-MM-DD HH:mm:ss!UTC} | {message}",
        ),
    ],
)



from functools import cached_property
from pathlib import Path

import pandas as pd

from .._io import loads_pickle
from .._models import BaseModel, create_model
from .._settings import check_dynamicdir_exists, settings

HERE = Path(__file__).parent
SPECIES_FILENAME = HERE / "tables/Species.csv"


SpeciesData = create_model("SpeciesData", __module__=__name__)


class Entry(BaseModel):
    display_name: str
    scientific_name: str
    common_name: str
    taxon_id: int
    assembly: str
    accession: str
    release: int
    short_name: str


class Species:
    """Species class."""

    def __init__(self, common_name="human") -> None:
        self._std_name = common_name
        self._dataclasspath = settings.dynamicdir / "speciesdata.pkl"

    @cached_property
    def df(self):
        self._df = pd.read_csv(SPECIES_FILENAME, header=0, index_col=0)
        return self._df

    @property
    def std_id(self):
        """common_name is the standardized id for species."""
        return "display_name"

    @property
    def std_name(self):
        """Value of the .std_id."""
        return self._std_name

    @property
    def fields(self):
        return self.df.columns.tolist()

    @property
    def dataclasspath(self):
        """Path to the picked dataclass."""
        return self._dataclasspath

    @cached_property
    def dataclass(self):
        return self._load_dataclass()

    def search(self, field: str):
        """Search species fields based on .std_id.

        Args:
            field: one of .fields
        Returns:
            value of a field

        e.g.
        'common_name': 'human'
        'scientific_name': 'homo_sapiens'
        'short_name': 'hsapiens'
        'taxon_id': 9606
        'assembly': 'GRCh38.p13'

        """
        return self.df[[field]].to_dict()[field][self.std_name]

    @check_dynamicdir_exists
    def _load_dataclass(self):
        """Pydantic data class of genes."""
        if not self.dataclasspath.exists():
            import pickle

            from .._io import write_pickle

            model = self._create_data_model()
            write_pickle(pickle.dumps(model()), self.dataclasspath)

        return loads_pickle(self.dataclasspath)

    def _create_data_model(self):
        """Create the species data model with pydantic."""
        df = self.df.reset_index()
        for i in df.index:
            entry = {}
            entry.update({col: df.loc[i][col] for col in df.columns})
            SpeciesData.add_fields(**{df.loc[i][self.std_id]: (Entry, Entry(**entry))})

        return SpeciesData



import pandas as pd

from .._servers import EnsemblREST
from .main import SPECIES_FILENAME

SPECIES_COLS = [
    "scientific_name",
    "display_name",
    "common_name",
    "taxon_id",
    "assembly",
    "accession",
    "release",
]


def update_species_table() -> None:
    """Fetch species table from Ensembl REST.

    Returns:
        a dataframe
    """
    entries = EnsemblREST().species_info()

    # format into a dataframe
    sp_dict: dict = {}
    cols = [
        "display_name",
        "common_name",
        "taxon_id",
        "assembly",
        "accession",
        "release",
    ]
    for entry in entries:
        id = entry.get("@name")
        sp_dict[id] = []
        for i in SPECIES_COLS:
            if i != "scientific_name":
                sp_dict[id].append(entry.get(f"@{i}"))
    sp_df = pd.DataFrame.from_dict(sp_dict).T
    sp_df.columns = cols
    sp_df.index.name = "scientific_name"
    sp_df["display_name"] = sp_df["display_name"].str.lower()
    # Adding a short_name column
    sp_df["short_name"] = [f'{i[0].lower()}{i.split("_")[-1]}' for i in sp_df.index]
    # Set display_name as the index for std_id
    sp_df = sp_df.reset_index().set_index("display_name")
    sp_df.to_csv(SPECIES_FILENAME, header=True, index=True)
